/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.wsdl.Binding;
import javax.wsdl.BindingOperation;
import javax.wsdl.Definition;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.xml.namespace.QName;

import org.apache.log4j.Logger;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.config.DefinitionCacheConfig;
import com.eviware.soapui.config.EndpointsConfig;
import com.eviware.soapui.config.InterfaceConfig;
import com.eviware.soapui.config.OperationConfig;
import com.eviware.soapui.config.SoapVersionTypesConfig;
import com.eviware.soapui.impl.wsdl.mock.WsdlMockResponse;
import com.eviware.soapui.impl.wsdl.support.soap.SoapMessageBuilder;
import com.eviware.soapui.impl.wsdl.support.soap.SoapVersion;
import com.eviware.soapui.impl.wsdl.support.wsdl.CachedWsdlLoader;
import com.eviware.soapui.impl.wsdl.support.wsdl.UrlWsdlLoader;
import com.eviware.soapui.impl.wsdl.support.wsdl.WsdlContext;
import com.eviware.soapui.impl.wsdl.support.wsdl.WsdlLoader;
import com.eviware.soapui.impl.wsdl.support.wsdl.WsdlUtils;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlTestRequest;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlTestRequestStep;
import com.eviware.soapui.model.ModelItem;
import com.eviware.soapui.model.iface.Interface;
import com.eviware.soapui.model.iface.InterfaceListener;
import com.eviware.soapui.model.iface.Operation;
import com.eviware.soapui.settings.WsdlSettings;
import com.eviware.soapui.support.UISupport;
import com.eviware.soapui.support.types.StringList;

/**
 * WSDL implementation of Interface, maps to a WSDL Binding
 * 
 * @author Ole.Matzura
 */

public class WsdlInterface extends AbstractWsdlModelItem<InterfaceConfig> implements Interface
{
	public static final String STYLE_DOCUMENT = "Document";
	public static final String STYLE_RPC = "RPC";

	public static final String JBOSSWS_ACTIONS = "jbossws";
	public static final String WSTOOLS_ACTIONS = "wstools";
	public static final String XML_ACTIONS = "xml";

	private final static Logger log = Logger.getLogger( WsdlInterface.class );
	private List<WsdlOperation> operations = new ArrayList<WsdlOperation>();
	private WsdlProject project;
	private SoapMessageBuilder soapMessageBuilder;
	private WsdlContext wsdlContext;
	private Set<InterfaceListener> interfaceListeners = new HashSet<InterfaceListener>();
	private boolean updating = false;

	public WsdlInterface( WsdlProject project, InterfaceConfig interfaceConfig )
	{
		super( interfaceConfig, project, "/interface2.gif" );

		this.project = project;

		if( interfaceConfig.getEndpoints() == null )
			interfaceConfig.addNewEndpoints();

		List<OperationConfig> operationConfigs = interfaceConfig.getOperationList();
		for( int i = 0; i < operationConfigs.size(); i++ )
		{
			operations.add( new WsdlOperation( this, operationConfigs.get( i ) ) );
		}

		for( InterfaceListener listener : SoapUI.getListenerRegistry().getListeners( InterfaceListener.class ) )
		{
			addInterfaceListener( listener );
		}
	}

	public String[] getEndpoints()
	{
		EndpointsConfig endpoints = getConfig().getEndpoints();
		List<String> endpointArray = endpoints.getEndpointList();
		return endpointArray.toArray( new String[endpointArray.size()] );
	}

	public WsdlOperation getOperationAt( int index )
	{
		return operations.get( index );
	}

	public int getOperationCount()
	{
		return operations.size();
	}
	
	public WsdlOperation addNewOperation( BindingOperation operation )
	{
		WsdlOperation operationImpl = new WsdlOperation( this, getConfig().addNewOperation() );
		operations.add( operationImpl );

		operationImpl.initFromBindingOperation( operation );
		fireOperationAdded( operationImpl );
		return operationImpl;
	}

	public WsdlProject getProject()
	{
		return project;
	}

	public void addEndpoint( String endpoint )
	{
		if( endpoint == null || endpoint.trim().length() == 0 )
			return;

		endpoint = endpoint.trim();
		String[] endpoints = getEndpoints();

		// dont add the same endpoint twice
		if( Arrays.asList( endpoints ).contains( endpoint ) )
			return;

		getConfig().getEndpoints().addNewEndpoint().setStringValue( endpoint );

		notifyPropertyChanged( ENDPOINT_PROPERTY, null, endpoint );
	}

	public void changeEndpoint( String oldEndpoint, String newEndpoint )
	{
		if( oldEndpoint == null || oldEndpoint.trim().length() == 0 )
			return;
		if( newEndpoint == null || newEndpoint.trim().length() == 0 )
			return;

		EndpointsConfig endpoints = getConfig().getEndpoints();

		for( int c = 0; c < endpoints.sizeOfEndpointArray(); c++ )
		{
			if( endpoints.getEndpointArray( c ).equals( oldEndpoint ) )
			{
				endpoints.setEndpointArray( c, newEndpoint );
				notifyPropertyChanged( ENDPOINT_PROPERTY, oldEndpoint, newEndpoint );
				break;
			}
		}
	}

	public void removeEndpoint( String endpoint )
	{
		EndpointsConfig endpoints = getConfig().getEndpoints();

		for( int c = 0; c < endpoints.sizeOfEndpointArray(); c++ )
		{
			if( endpoints.getEndpointArray( c ).equals( endpoint ) )
			{
				endpoints.removeEndpoint( c );
				notifyPropertyChanged( ENDPOINT_PROPERTY, endpoint, null );
				break;
			}
		}
	}

	public void setDefinition( String wsdlUrl, boolean cache )
	{
		String old = getDefinition();

		getConfig().setDefinition( wsdlUrl );

		if( wsdlContext != null )
		{
			wsdlContext.setDefinition( wsdlUrl, getConfig().getDefinitionCache() );
			wsdlContext.setSoapVersion( getSoapVersion() );
		}

		notifyPropertyChanged( DEFINITION_PROPERTY, old, wsdlUrl );
		
		notifyPropertyChanged( UPDATING_PROPERTY, true, false );
	}

	public DefinitionCacheConfig cacheDefinition( WsdlLoader loader ) throws Exception
	{
		log.debug( "Caching definition for [" + loader.getBaseURI() + "]" );
		if( getConfig().isSetDefinitionCache() )
			getConfig().unsetDefinitionCache();

		DefinitionCacheConfig definitionCache = getConfig().addNewDefinitionCache();
		definitionCache.set( WsdlLoader.cacheWsdl( loader ) );
		return definitionCache;
	}

	public String getDefinition()
	{
		return getConfig().isSetDefinition() ? getConfig().getDefinition() : null;
	}

	public synchronized WsdlContext getWsdlContext()
	{
		if( wsdlContext == null )
		{
			wsdlContext = new WsdlContext( getDefinition(), getSoapVersion(), getConfig().getDefinitionCache(), this );
		}

		return wsdlContext;
	}

	/**
	 * Used by importer so we dont need to reload the context after importing..
	 * 
	 * @param wsdlContext
	 */

	public void setWsdlContext( WsdlContext wsdlContext )
	{
		this.wsdlContext = wsdlContext;
		this.wsdlContext.setSoapVersion( getSoapVersion() );
		this.wsdlContext.setInterface( this );

		if( !getConfig().isSetDefinitionCache() )
			getConfig().addNewDefinitionCache();

		if( wsdlContext.getCacheConfig() != null )
		{
			// use cache from context
			getConfig().setDefinitionCache( wsdlContext.getCacheConfig() );
		}
	}

	public SoapMessageBuilder getMessageBuilder()
	{
		if( soapMessageBuilder == null )
		{
			try
			{
				soapMessageBuilder = new SoapMessageBuilder( this );
			}
			catch( Exception e )
			{
				SoapUI.logError( e );
			}
		}
		return soapMessageBuilder;
	}

	public void setSoapMessageBuilder( SoapMessageBuilder builder )
	{
		soapMessageBuilder = builder;
		soapMessageBuilder.setInterface( this );
	}

	public QName getBindingName()
	{
		return getConfig().getBindingName() == null ? null : QName.valueOf( getConfig().getBindingName() );
	}

	public void setBindingName( QName name )
	{
		getConfig().setBindingName( name.toString() );
	}

	public SoapVersion getSoapVersion()
	{
		if( getConfig().getSoapVersion() == SoapVersionTypesConfig.X_1_2 )
			return SoapVersion.Soap12;

		return SoapVersion.Soap11;
	}

	public void setSoapVersion( SoapVersion version )
	{
		if( version == SoapVersion.Soap11 )
			getConfig().setSoapVersion( SoapVersionTypesConfig.X_1_1 );
		else if( version == SoapVersion.Soap12 )
			getConfig().setSoapVersion( SoapVersionTypesConfig.X_1_2 );
		else
			throw new RuntimeException( "Unknown soapVersion [" + version + "], must be 1.1 or 1.2" );

		getWsdlContext().setSoapVersion( version );
	}

	public boolean updateDefinition( String url, boolean createRequests ) throws Exception
	{
		if( getConfig().isSetDefinitionCache() )
			getConfig().unsetDefinitionCache();

		WsdlContext newContext = getNewContext( url, getSettings().getBoolean( WsdlSettings.CACHE_WSDLS ) );
		if( !newContext.load() )
		{
			return false;
		}

		BindingTuple tuple = findBinding( newContext );
		if( tuple == null )
			return false;

		setBindingName( tuple.binding.getQName() );

		// update name
		if( getSettings().getBoolean( WsdlSettings.NAME_WITH_BINDING ) )
			setName( tuple.binding.getQName().getLocalPart() );

		// update context
		wsdlContext = newContext;
		if( soapMessageBuilder != null )
			soapMessageBuilder.setWsdlContext( wsdlContext );

		transferOperations( tuple.binding, createRequests );

		setDefinition( url, false );

		transferEndpoints( tuple.port );

		getProject().fireInterfaceUpdated( this );

		return true;
	}

	public BindingTuple prepareUpdateDefinition( String url ) throws Exception
	{
		WsdlContext newContext = getNewContext( url, false );
		if( !newContext.load() )
		{
			return null;
		}

		BindingTuple tuple = findBinding( newContext );
		return tuple;
	}

	public void updateDefinition( BindingTuple tuple ) throws Exception
	{
		setBindingName( tuple.binding.getQName() );

		if( getConfig().isSetDefinitionCache() )
			getConfig().unsetDefinitionCache();

		// update name
		if( getSettings().getBoolean( WsdlSettings.NAME_WITH_BINDING ) )
			setName( tuple.binding.getQName().getLocalPart() );

		// update context
		wsdlContext = tuple.context;
		if( soapMessageBuilder != null )
			soapMessageBuilder.setWsdlContext( wsdlContext );
	}

	private WsdlContext getNewContext( String url, boolean cache )
	{
		if( !cache )
		{
			return new WsdlContext( url, getSoapVersion(), null, null );
		}
		else
		{
			return new WsdlContext( url, getSoapVersion(), null, this );
		}
	}

	public BindingOperation findBindingOperation( Definition definition, String bindingOperationName, String inputName,
				String outputName )
	{
		Binding binding = definition.getBinding( getBindingName() );
		return WsdlUtils.findBindingOperation( binding, bindingOperationName, inputName, outputName );
	}

	@SuppressWarnings( "unchecked" )
	private BindingTuple findBinding( WsdlContext newContext ) throws Exception
	{
		BindingTuple tuple = new BindingTuple();
		tuple.context = newContext;

		// start by finding the old binding in the new definition
		Definition definition = newContext.getDefinition();
		Map serviceMap = definition.getAllServices();
		Iterator<String> i = serviceMap.keySet().iterator();
		while( i.hasNext() )
		{
			tuple.service = ( Service ) serviceMap.get( i.next() );
			Map portMap = tuple.service.getPorts();

			Iterator i2 = portMap.keySet().iterator();
			while( i2.hasNext() )
			{
				tuple.port = ( Port ) portMap.get( i2.next() );
				if( tuple.port.getBinding().getQName().equals( getBindingName() ) )
				{
					tuple.binding = tuple.port.getBinding();
				}
			}

			if( tuple.binding != null )
				break;
			tuple.service = null;
		}

		if( tuple.service == null && tuple.binding == null )
		{
			tuple.binding = definition.getBinding( getBindingName() );
		}

		// missing matching binding, prompt for new one to use instead (will
		// happen if binding has been renamed)
		if( tuple.binding == null )
		{
			Map bindings = definition.getAllBindings();

			Object retval = UISupport.prompt( "Missing matching binding [" + getBindingName()
						+ "] in definition, select new\nbinding to map to", "Map Binding", bindings.keySet().toArray() );

			if( retval == null )
				return null;

			tuple.binding = ( Binding ) bindings.get( retval );
		}

		return tuple;
	}

	@SuppressWarnings( "unchecked" )
	public void transferOperations( Binding binding, boolean createRequests )
	{
		// prepare for transfer of operations/requests
		List<BindingOperation> newOperations = new ArrayList<BindingOperation>( binding.getBindingOperations() );
		Map<String, WsdlOperation> oldOperations = new HashMap<String, WsdlOperation>();
		for( int c = 0; c < operations.size(); c++ )
			oldOperations.put( operations.get( c ).getBindingOperationName(), operations.get( c ) );

		// clear existing from both collections
		for( int c = 0; c < newOperations.size(); c++ )
		{
			BindingOperation newOperation = newOperations.get( c );
			String bindingOperationName = newOperation.getName();
			if( oldOperations.containsKey( bindingOperationName ) )
			{
				log.info( "Synchronizing existing operation [" + bindingOperationName + "]" );
				WsdlOperation wsdlOperation = oldOperations.get( bindingOperationName );
				wsdlOperation.initFromBindingOperation( newOperation );
				fireOperationUpdated( wsdlOperation );

				oldOperations.remove( bindingOperationName );
				newOperations.remove( c );
				c--;
			}
		}

		// remove leftover operations
		Iterator<String> i = oldOperations.keySet().iterator();
		while( i.hasNext() )
		{
			String name = i.next();

			if( newOperations.size() > 0 )
			{
				List<String> list = new ArrayList<String>();
				list.add( "none - delete operation" );
				for( int c = 0; c < newOperations.size(); c++ )
					list.add( newOperations.get( c ).getName() );

				String retval = ( String ) UISupport.prompt( "Binding operation [" + name
							+ "] not found in new interface, select new\nbinding operation to map to", "Map Operation", list
							.toArray(), "none/cancel - delete operation" );

				int ix = retval == null ? -1 : list.indexOf( retval ) - 1;

				// delete operation?
				if( ix < 0 )
				{
					deleteOperation( name );
				}
				// change operation?
				else
				{
					BindingOperation newOperation = newOperations.get( ix );
					WsdlOperation wsdlOperation = oldOperations.get( name );
					wsdlOperation.initFromBindingOperation( newOperation );
					fireOperationUpdated( wsdlOperation );
					newOperations.remove( ix );
				}

				oldOperations.remove( name );
			}
			else
			{
				deleteOperation( name );
				oldOperations.remove( name );
			}

			i = oldOperations.keySet().iterator();
		}

		// add leftover new operations
		if( newOperations.size() > 0 )
		{
			for( int c = 0; c < newOperations.size(); c++ )
			{
				BindingOperation newOperation = newOperations.get( c );
				WsdlOperation wsdlOperation = addNewOperation( newOperation );

				if( createRequests )
				{
					WsdlRequest request = wsdlOperation.addNewRequest( "Request 1" );
					try
					{
						request.setRequestContent( wsdlOperation.createRequest( true ) );
					}
					catch( Exception e )
					{
						SoapUI.logError( e );
					}
				}
			}
		}
	}

	public void transferEndpoints( Port port )
	{
		if( port != null )
		{
			String endpoint = WsdlUtils.getSoapEndpoint( port );
			if( endpoint != null )
			{
				StringList list = new StringList( getEndpoints() );
				if( !list.contains( endpoint ) )
				{
					if( UISupport.confirm( "Update existing requests with new endpoint\n[" + endpoint + "]",
								"Update Definition" ) )
					{
						for( int c = 0; c < getOperationCount(); c++ )
						{
							Operation operation = getOperationAt( c );

							for( int ix = 0; ix < operation.getRequestCount(); ix++ )
							{
								operation.getRequestAt( ix ).setEndpoint( endpoint );
							}
						}
					}

					addEndpoint( endpoint );
				}
			}
		}
	}

	public void deleteOperation( String bindingOperationName )
	{
		for( int c = 0; c < operations.size(); c++ )
		{
			WsdlOperation wsdlOperation = operations.get( c );
			if( wsdlOperation.getBindingOperationName().equals( bindingOperationName ) )
			{
				log.info( "deleting operation [" + bindingOperationName + "]" );

				// remove requests first (should this be done by some listener?)
				while( wsdlOperation.getRequestCount() > 0 )
					wsdlOperation.removeRequest( wsdlOperation.getRequestAt( 0 ) );

				operations.remove( c );

				try
				{
					fireOperationRemoved( wsdlOperation );
				}
				finally
				{
					wsdlOperation.release();
					getConfig().removeOperation( c );
				}

				return;
			}
		}
	}

	public void fireOperationAdded( WsdlOperation operation )
	{
		InterfaceListener[] a = interfaceListeners.toArray( new InterfaceListener[interfaceListeners.size()] );

		for( int c = 0; c < a.length; c++ )
		{
			a[c].operationAdded( operation );
		}
	}

	public void fireOperationUpdated( WsdlOperation operation )
	{
		InterfaceListener[] a = interfaceListeners.toArray( new InterfaceListener[interfaceListeners.size()] );

		for( int c = 0; c < a.length; c++ )
		{
			a[c].operationUpdated( operation );
		}
	}

	public void fireOperationRemoved( WsdlOperation operation )
	{
		InterfaceListener[] a = interfaceListeners.toArray( new InterfaceListener[interfaceListeners.size()] );

		for( int c = 0; c < a.length; c++ )
		{
			a[c].operationRemoved( operation );
		}
	}

	public void fireRequestAdded( WsdlRequest request )
	{
		InterfaceListener[] a = interfaceListeners.toArray( new InterfaceListener[interfaceListeners.size()] );

		for( int c = 0; c < a.length; c++ )
		{
			a[c].requestAdded( request );
		}
	}

	public void fireRequestRemoved( WsdlRequest request )
	{
		InterfaceListener[] a = interfaceListeners.toArray( new InterfaceListener[interfaceListeners.size()] );

		for( int c = 0; c < a.length; c++ )
		{
			a[c].requestRemoved( request );
		}
	}

	public void addInterfaceListener( InterfaceListener listener )
	{
		interfaceListeners.add( listener );
	}

	public void removeInterfaceListener( InterfaceListener listener )
	{
		interfaceListeners.remove( listener );
	}

	public WsdlOperation getOperationByName( String name )
	{
		return ( WsdlOperation ) getWsdlModelItemByName( operations, name );
	}

	public boolean isCached()
	{
		return getConfig().isSetDefinitionCache();
	}

	public WsdlLoader createWsdlLoader()
	{
		return isCached() ? new CachedWsdlLoader( getConfig().getDefinitionCache() )
					: new UrlWsdlLoader( getDefinition() );
	}

	public void clearCache()
	{
		if( wsdlContext != null )
			wsdlContext.setDefinitionCache( null );

		if( getConfig().isSetDefinitionCache() )
			getConfig().unsetDefinitionCache();
	}

	public String getStyle()
	{
		if( wsdlContext == null || !wsdlContext.isLoaded() )
			return "<not loaded>";

		try
		{
			Binding binding = wsdlContext.getDefinition().getBinding( getBindingName() );
			if( binding == null )
				return "<missing binding>";

			if( WsdlUtils.isRpc( binding ) )
			{
				return STYLE_RPC;
			}
			else
			{
				return STYLE_DOCUMENT;
			}
		}
		catch( Exception e )
		{
			SoapUI.logError( e );
			return "<error>";
		}
	}

	@Override
	public void release()
	{
		super.release();

		for( WsdlOperation operation : operations )
			operation.release();

		interfaceListeners.clear();
	}

	public List<Operation> getOperationList()
	{
		return new ArrayList<Operation>( operations );
	}

	@Override
	public void beforeSave()
	{
		for( WsdlOperation operation : operations )
			operation.beforeSave();
	}

	public static class BindingTuple
	{
		public WsdlContext context = null;
		public Service service = null;
		public Port port = null;
		public Binding binding = null;
	}

	public List<? extends ModelItem> getChildren()
	{
		return getOperationList();
	}

	public boolean isUpdating()
	{
		return updating;
	}

	public void setUpdating( boolean updating )
	{
		if( this.updating == updating )
			return;

		if( updating )
		{
			List<AbstractWsdlModelItem> messages = getAllMessages();
			for( AbstractWsdlModelItem modelItem : messages )
			{
				modelItem.beforeSave();
			}
		}

		boolean oldValue = this.updating;
		this.updating = updating;

		notifyPropertyChanged( UPDATING_PROPERTY, oldValue, updating );
	}

	public List<AbstractWsdlModelItem> getAllMessages()
	{
		ArrayList<AbstractWsdlModelItem> list = new ArrayList<AbstractWsdlModelItem>();
		getAllMessages( getProject(), list );
		return list;
	}

	private void getAllMessages( ModelItem modelItem, List<AbstractWsdlModelItem> list )
	{
		if( modelItem instanceof WsdlRequest )
		{
			WsdlRequest wsdlRequest = ( WsdlRequest ) modelItem;
			if( wsdlRequest.getOperation().getInterface() == this )
				list.add( wsdlRequest );
		}
		else if( modelItem instanceof WsdlTestRequestStep )
		{
			WsdlTestRequestStep testRequestStep = ( WsdlTestRequestStep ) modelItem;
			WsdlTestRequest testRequest = testRequestStep.getTestRequest();
			if( testRequest.getOperation().getInterface() == this )
				list.add( testRequest );
		}
		else if( modelItem instanceof WsdlMockResponse )
		{
			WsdlMockResponse mockResponse = ( WsdlMockResponse ) modelItem;
			if( mockResponse.getMockOperation().getOperation().getInterface() == this )
				list.add( mockResponse );
		}

		// Traverse the ModelItem hierarchy.
		for( ModelItem child : modelItem.getChildren() )
		{
			getAllMessages( child, list );
		}
	}
}
