/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.support;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.config.PropertiesTypeConfig;
import com.eviware.soapui.config.PropertyConfig;
import com.eviware.soapui.impl.wsdl.MutableTestPropertyHolder;
import com.eviware.soapui.impl.wsdl.support.wsdl.UrlWsdlLoader;
import com.eviware.soapui.model.ModelItem;
import com.eviware.soapui.model.propertyexpansion.PropertyExpansion;
import com.eviware.soapui.model.testsuite.TestProperty;
import com.eviware.soapui.model.testsuite.TestPropertyListener;

public class XmlBeansPropertiesTestPropertyHolder implements MutableTestPropertyHolder
{
	private PropertiesTypeConfig config;
	private List<PropertiesStepProperty> properties = new ArrayList<PropertiesStepProperty>();
	private Map<String,PropertiesStepProperty> propertyMap = new HashMap<String, PropertiesStepProperty>();
	private Set<TestPropertyListener> listeners = new HashSet<TestPropertyListener>();
	public ModelItem modelItem;

	public XmlBeansPropertiesTestPropertyHolder( ModelItem modelItem, PropertiesTypeConfig config )
	{
		this.modelItem = modelItem;
		this.config = config;
		
		for( PropertyConfig propertyConfig : config.getPropertyList())
		{
			addProperty( propertyConfig, false );
		}
	}

	private PropertiesStepProperty addProperty( PropertyConfig propertyConfig, boolean notify )
	{
		PropertiesStepProperty propertiesStepProperty = new PropertiesStepProperty( propertyConfig );
		properties.add( propertiesStepProperty );
		propertyMap.put( propertiesStepProperty.getName().toUpperCase(), propertiesStepProperty );
		
		if( notify )
		{
			firePropertyAdded( propertiesStepProperty.getName() );
		}
		
		return propertiesStepProperty;
	}
	
	private void firePropertyAdded( String name )
	{
		TestPropertyListener [] listenersArray = listeners.toArray( new TestPropertyListener[listeners.size()] );
		for( TestPropertyListener listener : listenersArray )
		{
			listener.propertyAdded( name );
		}
	}
	
	private void firePropertyRemoved( String name )
	{
		TestPropertyListener [] listenersArray = listeners.toArray( new TestPropertyListener[listeners.size()] );
		for( TestPropertyListener listener : listenersArray )
		{
			listener.propertyRemoved( name );
		}
	}
	
	private void firePropertyRenamed( String oldName, String newName )
	{
		TestPropertyListener [] listenersArray = listeners.toArray( new TestPropertyListener[listeners.size()] );
		for( TestPropertyListener listener : listenersArray )
		{
			listener.propertyRenamed( oldName, newName );
		}
	}
	
	private void firePropertyValueChanged( String name, String oldValue, String newValue )
	{
		TestPropertyListener [] listenersArray = listeners.toArray( new TestPropertyListener[listeners.size()] );
		for( TestPropertyListener listener : listenersArray )
		{
			listener.propertyValueChanged(name, oldValue, newValue );
		}
	}

	public TestProperty addProperty( String name )
	{
		PropertyConfig propertyConfig = config.addNewProperty();
		propertyConfig.setName( name );
		return addProperty( propertyConfig, true );
	}

	public void addTestPropertyListener( TestPropertyListener listener )
	{
		listeners.add( listener );
	}

	public PropertiesStepProperty getProperty( String name )
	{
		return propertyMap.get( name.toUpperCase() );
	}

	public String[] getPropertyNames()
	{
		String [] result = new String[properties.size()];
		for( int c = 0; c < properties.size(); c++ )
			result[c] = properties.get( c ).getName();
		
		return result;
	}

	public String getPropertyValue( String name )
	{
		TestProperty property = getProperty( name );
		return property == null ? null : property.getValue();
	}

	public void removeProperty( String propertyName )
	{
		TestProperty property = getProperty( propertyName );
		if( property != null )
		{
			int ix = properties.indexOf( property );
			propertyMap.remove( propertyName.toUpperCase() );
			properties.remove( ix );
			config.removeProperty( ix );
			
			firePropertyRemoved( propertyName );
		}
	}

	public void removeTestPropertyListener( TestPropertyListener listener )
	{
		listeners.remove( listener );
	}

	public void setPropertyValue( String name, String value )
	{
		PropertiesStepProperty property = getProperty( name );
		if( property != null )
			property.setValue( value );
		else
			addProperty( name ).setValue( value );
	}

	public void resetPropertiesConfig( PropertiesTypeConfig config )
	{
		this.config = config;
		
		for( int c = 0; c < config.sizeOfPropertyArray(); c++ )
		{
			properties.get( c ).setConfig( config.getPropertyArray( c ));
		}	
		
	}

	public boolean renameProperty( String name, String newName )
	{
		if( getProperty( newName ) != null )
			return false;
		
		PropertiesStepProperty property = getProperty( name );
		if( property == null )
			return false;
		
		property.setName( newName );
		return true;
	}

	/**
	 * Internal property class
	 * 
	 * @author ole
	 */
	
	public class PropertiesStepProperty implements TestProperty
	{
		private PropertyConfig propertyConfig;

		public PropertiesStepProperty(PropertyConfig propertyConfig)
		{
			this.propertyConfig = propertyConfig;
		}

		public void setConfig(PropertyConfig propertyConfig)
		{
			this.propertyConfig = propertyConfig;
		}

		public String getName()
		{
			return propertyConfig.getName();
		}
		
		public void setName( String name )
		{
			String oldName = getName();
			propertyConfig.setName( name );
			
			propertyMap.remove( oldName.toUpperCase() );
			propertyMap.put( name.toUpperCase(), this );
			
			firePropertyRenamed( oldName, name );
		}

		public String getDescription()
		{
			return null;
		}

		public String getValue()
		{
			return propertyConfig.getValue();
		}

		public void setValue(String value)
		{
			String oldValue = getValue();
			propertyConfig.setValue( value );
			
			firePropertyValueChanged( getName(), oldValue, value );
		}

		public boolean isReadOnly()
		{
			return false;
		}

		public Type getType()
		{
			return Type.STRING;
		}

		public ModelItem getModelItem()
		{
			return modelItem;
		}
	}

	public void saveTo( Properties props )
	{
		int cnt = 0;
		for( PropertiesStepProperty p : properties )
		{
			String name = p.getName();
			String value = p.getValue();
			if( value == null )
				value = "";
			
			props.setProperty( name, value );
			cnt++;
		}
	}

	public int getPropertyCount()
	{
		return properties.size();
	}

	public TestProperty getPropertyAt( int index )
	{
		return properties.get( index );
	}

	public Map<String, TestProperty> getProperties()
	{
		Map<String,TestProperty> result = new HashMap<String,TestProperty>();
		for( String name : propertyMap.keySet() )
			result.put( name, propertyMap.get( name ));
		
		return result;
	}

	public boolean hasProperty( String name )
	{
		return propertyMap.containsKey( name.toUpperCase() );
	}

	public int addPropertiesFromFile( String propFile )
	{
		try
		{
			InputStream input = null;
			
			File file = new File( propFile );
			if( file.exists() )
			{
				input = new FileInputStream( file );
			}
			else if( propFile.toLowerCase().startsWith( "http://" ) || propFile.toLowerCase().startsWith( "https://" ))
			{
			   UrlWsdlLoader loader = new UrlWsdlLoader( propFile );
			   loader.setUseWorker( false );
			   input = loader.load();
			}

			Properties properties = new Properties();
			properties.load( input );
			
			for( Object key : properties.keySet() )
			{
				String name = key.toString();
				if( !hasProperty( name ))
					addProperty( name ).setValue( properties.getProperty( name ) );
				else
					setPropertyValue( name, properties.getProperty( name ) );
			}
			
			return properties.size();
		}
		catch( Exception e )
		{
			SoapUI.logError( e );
		}
		
		return 0;
	}

	public ModelItem getModelItem()
	{
		return modelItem;
	}

	public PropertyExpansion[] getPropertyExpansions()
	{
		List<PropertyExpansion> result = new ArrayList<PropertyExpansion>();
		
		return result.toArray( new PropertyExpansion[result.size()] );
	}
}
